/** Data of a table scheme definition */
class SchemeDefTable {
    constructor(
        table   // Name of the table described
    ) {
        this.table = table;
        this.attributes = [];
        let id = -1;
        this.createReferenceGroupId = () => ++id;
    }

    /**
    * Create id of new reference group
    * @return Number    Created reference group's id
    */
    createReferenceGroupId() {
        return NaN;
    }

    /**
    * Get information on reference group with given id
    * @return Object    { table: "Name of referenced table", mapping: "Array of objects like { attribute, referencing }" }
    */
    getReferenceInfo(
        id      // Id of reference group
    ) {
        const impactedAttributes = this.attributes.filter(att => att.reference && att.reference.group == id);
        if (!impactedAttributes.length)
            return null;
        return { table: impactedAttributes[0].reference.table, mapping: impactedAttributes.map(att => ({ attribute: att.name, referencing: att.reference.attribute })) };
    }

    /**
    * Get information on all reference groups of this table scheme definition
    * @return Array     Array of return values of getReferenceInfo corresponding to array's index
    */
    getAllReferenceInfos() {
        const maxGroupId = this.attributes.reduce((prev, curr) => Math.max(prev, curr.reference ? curr.reference.group : 0), 0);
        return (new Array(maxGroupId + 1)).fill(0).map((value, index) => this.getReferenceInfo(index));
    }

    /**
    * Add attribute to table
    * @return undefined
    */
    addAttribute(
        attribute       // SchemeDefAttribute instance to add
    ) {
        this.attributes.push(attribute);
    }

    /**
    * Get contained SchemeDefAttribute instance having name `name`
    * @return SchemeDefAttribute    Requested attribute
    */
    getAttributeByName(
        name        // Name of requested attribute
    ) {
        return this.attributes.reduce((prev, attribute) => prev || ((attribute.name == name) ? attribute : null), null);
    }

    /**
    * Find name of first duplicate attribute
    * @return mixed         First found attribute name given to at least two attributes
    */
    getDuplicateName() {
        const names = [];
        return this.attributes.reduce((prev, { name }) => {
            if (prev)
                return prev;
            if (names.indexOf(name) > -1)
                return name;
            names.push(name);
            return false;
        }, false);
    }

    /**
    * Check if there is a primary key
    * @return boolean       Is there a primary key?
    */
    hasPrimaryKey() {
        return this.attributes.reduce((prev, attribute) => prev || attribute.isPrimary, false);
    }

    /**
    * Check if all attributes declare a datatype
    * @return mixed     `true` if all attributes have a datatype declared, `false` if no attribute has a datatype declared and `undefined` otherwise or if there is no attribute
    */
    hasDatatypes() {
        if (!this.attributes.length)
            return undefined;
        if (!this.attributes.reduce((prev, attribute) => prev + ((attribute.datatype === undefined) ? 1 : 0), 0))
            return true;
        if (!this.attributes.reduce((prev, attribute) => prev + ((attribute.datatype === undefined) ? 0 : 1), 0))
            return false;
        return undefined;
    }

    /**
    * Return an error message if there is a reference group referencing attributes of different tables
    * @return mixed         Error message or `false` if there is no such error
    */
    checkReferences() {
        const tables = {};
        const firstAttribute = {};
        return this.attributes.reduce((prev, { name, reference }) => {
            if (prev || !(reference instanceof SchemeDefReference))
                return prev;
            const { group, table } = reference;
            if (tables[group] == undefined) {
                tables[group] = table;
                firstAttribute[group] = name;
                return prev;
            }
            if (tables[group] != table)
                return 'Die Fremdschlüsselattribute „' + firstAttribute[group] + '“ und „' + name + '“ von „' + this.table + '“ müssten sich auf Attribute derselben Tabelle beziehen.';
            return prev;
        }, false);
    }

    /**
    * Check if table scheme definition is semanticaly correct
    * @return mixed         Not falsey string describing first error or `false` (meaning no error)
    */
    validate(
        defs,                   // SchemeDefs instance containing this object
        arrAllowedDatatypes     // Array of strings representing all allowed data types
    ) {
        const duplicate = this.getDuplicateName();
        if (duplicate)
            return 'In der Tabelle „' + this.table + '“ darf es nur ein Attribut namens „' + duplicate + '“ geben.';
        if (!this.hasPrimaryKey())
            return 'In der Tabelle „' + this.table + '“ müsste es einen Primärschlüssel geben.';
        if (this.hasDatatypes() === undefined)
            return 'In der Tabelle „' + this.table + '“ müssten Datentypen immer oder niemals angegeben werden.';
        return this.checkReferences() || this.attributes.reduce((prev, attribute) => prev || attribute.validate(this, defs, arrAllowedDatatypes), false);
    }

    /**
    * Calculate best fit of this table scheme and target data
    * @return Object    Objects like { match: "Fraction of match", desc: "Human readable text describing main difference" }
    */
    match(
        targetData      // Description of wanted structure beeing object optionally having properties `table` (required table name) and/or `attributes` (required attributes)
    ) {
        const arrReferenceInfo = this.getAllReferenceInfos();
        if (!(targetData instanceof Object))
            return { match: 0, desc: 'Interner Fehler: Format der Tabellenschemavorgabe ist ungültig' };
        if (targetData.table === undefined && targetData.attributes === undefined)
            return { match: 1, desc: '' };
        const nameMismatch = (this.table != targetData.table);
        if (targetData.attributes === undefined)
            return { match: nameMismatch ? 0 : 1, desc: nameMismatch ? ('Tabelle „' + this.table + '“ müsste „' + targetData.table + '“ heißen') : '' };
        const attributesMatch = SchemeDefs.matchGeneral(
            this.attributes,
            (attr, target) => this.matchAttribute(attr, target, arrReferenceInfo),
            targetData.attributes,
            {
                internalTargetInvalid: 'Interner Fehler: Beschreibung der geforderten Attribute liegt in falschem Format vor',
                missingItem: target => (target.name ? ('Attribut „' + target.name + '“') : 'Attribut') + ' fehlt in der Tabelle „' + this.table + '“',
                surplusItem: item => 'Attribut „' + item.name + '“ in der Tabelle „' + this.table + '“ ist überflüssig',
                needMore: 'In der Tabelle „' + this.table + '“ müsste es mehr Attribute geben',
                needLess: 'In der Tabelle „' + this.table + '“ müsste es weniger Attribute geben'
            }
        );
        if (targetData.table === undefined)
            return attributesMatch;
        attributesMatch.match = (attributesMatch.match * Math.max(this.attributes.length, 1) + 1 - nameMismatch) / (Math.max(this.attributes.length, 1) + 1);
        if (nameMismatch)
            attributesMatch.desc = attributesMatch.desc || ('Tabelle „' + this.table + '“ müsste „' + targetData.table + '“ heißen');
        return attributesMatch;
    }

    /**
    * Calculate match of specific attribute to specific target data
    * @return Object    Objects like { match: "Fraction of match", desc: "Human readable text describing main difference" }
    */
    matchAttribute(
        attr,               // Attribute to match
        targetData,         // Object describing wanted properties of attribute
        arrReferenceInfo    // Optional: Return value of `getAllReferenceInfos`
    ) {
        let countFit = 0;
        let countTests = 0;
        let desc = '';
        if (!attr)
            return { match: 0, desc: 'Interner Fehler: Nicht existentes Attribut abgefragt' };
        if (targetData.name !== undefined) {
            ++countTests;
            if (targetData.name != attr.name)
                desc = desc || 'Der Name des Attributs „' + attr.name + '” in der Tabelle „' + this.table + '“ ist falsch';
            else
                ++countFit;
        }
        if (targetData.datatype !== undefined) {
            ++countTests;
            if (targetData.datatype != attr.datatype)
                desc = desc || 'Der Datentyp des Attributs „' + attr.name + '” in der Tabelle „' + this.table + '“ ist falsch';
            else
                ++countFit;
        }
        if (targetData.isPrimary !== undefined) {
            ++countTests;
            if (targetData.isPrimary != attr.isPrimary)
                desc = desc || 'Das Attribut „' + attr.name + '” in der Tabelle „' + this.table + (targetData.isPrimary ? '“ muss' : '“ darf nicht') + ' Teil des Primärschlüssels sein';
            else
                ++countFit;
        }
        if (targetData.reference !== undefined && !(targetData.reference instanceof Object)) {
            ++countTests;
            if (attr.reference)
                desc = desc || 'Das Attribut „' + attr.name + '” in der Tabelle „' + this.table + '“ kann kein Fremdschlüsselattribut sein';
            else
                ++countFit;
        }
        if (targetData.reference !== undefined && (targetData.reference instanceof Object) && !attr.reference) {
            ++countTests;
            desc = desc || 'Das Attribut „' + attr.name + '” in der Tabelle „' + this.table + '“ muss ein Fremdschlüsselattribut sein';
        }
        if (targetData.reference !== undefined && (targetData.reference instanceof Object) && attr.reference) {
            const groupId = Number.isFinite(attr.reference.group) ? attr.reference.group : -1;
            const referenceInfo = (arrReferenceInfo instanceof Array) ? arrReferenceInfo[groupId] : this.getReferenceInfo(groupId);
            const referenceMatch = this.matchReferenceInfo(targetData.reference, referenceInfo, attr);
            if (!desc && referenceMatch.desc)
                desc = referenceMatch.desc;
            countFit += referenceMatch.match;
            ++countTests;
        }
        return { match: countTests ? (countFit / countTests) : 1, desc };
    }

    /**
    */
    matchReferenceInfo(
        targetData,     // Object optionally having properties `table` (required value of referenced table) and `mapping` (array of objects having `attribute` and `referencing` properties) describing required data; `attribute === true` means this has to be the reference of attribute `attr`
        actualData,     // Object returned from SchemeDefTable.prototype.getReferenceInfo describing actual reference
        attr            // Attribute in question
    ) {
        if (!(actualData instanceof Object))
            return { match: 0, desc: 'Interner Fehler bei der Referenzprüfung des Attributs „' + attr.name + '” in der Tabelle „' + this.table + '“' };
        let countTests = 0;
        let countFit = 0;
        let desc = '';
        if (targetData.table !== undefined) {
            ++countTests;
            if (targetData.table != actualData.table)
                desc = desc || 'Das Attribut „' + attr.name + '” in der Tabelle „' + this.table + '“ muss ein die Tabelle „' + targetData.table + '“ referenzierendes Fremdschlüsselattribut sein';
            else
                ++countFit;
        }
        if (!(targetData.mapping instanceof Array))
            return { match: countTests ? (countFit / countTests) : 1, desc };
        const matrix = actualData.mapping.map(actual => targetData.mapping.map(target => {
            let countTests = 0;
            let countFit = 0;
            if (target.attribute !== undefined) {
                ++countTests;
                countFit += (((target.attribute === true) ? attr.name : target.attribute) == actual.attribute) ? 1 : 0;
            }
            if (target.referencing !== undefined) {
                ++countTests;
                countFit += (target.referencing == actual.referencing) ? 1 : 0;
            }
            return countTests ? (countFit / countTests) : 1;
        }));
        while (matrix.length) {
            const highestValue = matrix.reduce((prev, line) => Math.max(prev, line.reduce((p, v) => Math.max(p, v), 0), 0), 0);
            const occurencesByLine = matrix.map(line => line.filter(v => v == highestValue).length);
            const minOccurence = occurencesByLine.reduce((prev, curr) => (curr && curr < prev) ? curr : prev, Infinity);
            const actualIndex = occurencesByLine.reduce((prev, cur, index) => (prev != -1 || cur != minOccurence) ? prev : index, -1);
            if (actualIndex == -1)
                break;
            const targetIndex = matrix[actualIndex].reduce((prev, cur, index) => (prev != -1 || cur != highestValue) ? prev : index, -1);
            if (targetIndex == -1)
                break;
            ++countTests;
            if (highestValue != 1)
                desc = desc || 'Beim Attribut „' + attr.name + '” in der Tabelle „' + this.table + '“ ist die Referenz fehlerhaft';
            countFit += highestValue;
            matrix.splice(actualIndex, 1);
            matrix.forEach(line => line.splice(targetIndex, 1));
        }
        if (targetData.mapping.length != actualData.mapping.length) {
            countTests += Math.abs(targetData.mapping.length - actualData.mapping.length);
            desc = desc || 'Beim Attribut „' + attr.name + '” in der Tabelle „' + this.table + '“ ist die Referenz fehlerhaft';
        }
        return { match: countTests ? (countFit / countTests) : 1, desc };
    }
}

/** Name of table */
SchemeDefTable.prototype.table = '';

/** Contained attributes (SchemeDefAttribute instances) */
SchemeDefTable.prototype.attributes = [];



/**
* Read a single table scheme definition from tokens
* @return boolean       Found a table scheme definition?
*/
SchemeDefTable.parse = function(
    tokens,     // Array of tokens (without whitespace!) to read from
    schemeDefs  // Array of SchemeDefTable instances to write to
) {
    if (!tokens.length || tokens[0].type != SchemeDefToken.TYPE_IDENTIFIER)
        return false;
    const table = new SchemeDefTable(tokens.shift().content);
    schemeDefs.push(table);
    if (!tokens.length || tokens[0].content != '(')
        throw new SchemeDefParseError(tokens[0], 'öffnende Klammer');
    tokens.shift();

    let isPrimary = false;
    function parsePrimarySwitches() {
        while (tokens.length && tokens[0].content == '_') {
            isPrimary = !isPrimary;
            tokens.shift();
        }
    }

    function parseReference(attributename, groupId) {
        if (!tokens.length || tokens[0].type != SchemeDefToken.TYPE_IDENTIFIER)
            throw new SchemeDefParseError(tokens[0], 'Tabellenname');
        const reftable = tokens.shift().content;
        if (!tokens.length || tokens[0].content != '.')
            return new SchemeDefReference((groupId === undefined) ? table.createReferenceGroupId() : groupId, reftable, attributename);
        tokens.shift();
        if (!tokens.length || tokens[0].type != SchemeDefToken.TYPE_IDENTIFIER)
            throw new SchemeDefParseError(tokens[0], 'Attributname');
        return new SchemeDefReference((groupId === undefined) ? table.createReferenceGroupId() : groupId, reftable, tokens.shift().content);
    }

    function parseSimpleAttribute(disallowReference) {
        if (!tokens.length || tokens[0].type != SchemeDefToken.TYPE_IDENTIFIER)
            return false;
        const attribute = tokens.shift().content;
        let reference = null;
        if (tokens.length && tokens[0].content == '[') {
            if (disallowReference)
                throw new SchemeDefParseError(tokens[0], ['Doppelpunkt', 'Komma', 'schließende runde Klammer'])
            tokens.shift();
            reference = parseReference(attribute);
            if (!tokens.length || tokens[0].content != ']')
                throw new SchemeDefParseError(tokens[0], 'schließende eckige Klammer');
            tokens.shift();
        }
        let datatype = undefined;
        if (tokens.length && tokens[0].content == ':') {
            tokens.shift();
            if (!tokens.length || tokens[0].type != SchemeDefToken.TYPE_IDENTIFIER)
                throw new SchemeDefParseError(tokens[0], 'Datentyp');
            datatype = tokens.shift().content;
        }
        table.addAttribute(new SchemeDefAttribute(attribute, datatype, isPrimary, reference));
        return true;
    }

    function parseForeignKeyAttributes() {
        if (!tokens.length || tokens[0].content != '(')
            return false;
        tokens.shift();
        if (!tokens.length)
            throw new SchemeDefParseError(tokens[0], ['Attributname', 'Unterstreichungsumschalter']);
        const attributes = [];
        while (tokens.length) {
            parsePrimarySwitches();
            if (!parseSimpleAttribute(true))
                throw new SchemeDefParseError(tokens[0], 'Attributname');
            attributes.push(table.attributes[table.attributes.length - 1]);
            parsePrimarySwitches();
            if (tokens.length && tokens[0].content == ')') {
                tokens.shift();
                break;
            }
            if (tokens.length && tokens[0].content == ',') {
                tokens.shift();
                continue;
            }
            throw new SchemeDefParseError(tokens[0], ['Komma', 'Unterstreichungsumschalter', 'schließende runde Klammer']);
        }
        if (!tokens.length || tokens[0].content != '[')
            throw new SchemeDefParseError(tokens[0], 'öffnende eckige Klammer');
        tokens.shift();
        const groupId = table.createReferenceGroupId();
        if (tokens.length >= 2 && tokens[1].content == ']' && tokens[0].type == SchemeDefToken.TYPE_IDENTIFIER) {
            const reftable = tokens.shift().content;
            tokens.shift();
            attributes.forEach(attribute => attribute.reference = new SchemeDefReference(groupId, reftable, attribute.name));
            return true;
        }
        while (attributes.length) {
            attributes[0].reference = parseReference(attributes[0].name, groupId);
            attributes.shift();
            if (!tokens.length || tokens[0].content != (attributes.length ? ',' : ']'))
                throw new SchemeDefParseError(tokens[0], attributes.length ? 'Komma' : 'schließende eckige Klammer');
            tokens.shift();
        }
        return true;
    }

    while (tokens.length) {
        parsePrimarySwitches();
        if (!parseSimpleAttribute(false) && !parseForeignKeyAttributes())
            throw new SchemeDefParseError(tokens[0], ['Attributname', 'öffnende runde Klammer']);
        parsePrimarySwitches();
        if (tokens.length && tokens[0].content == ')') {
            tokens.shift();
            return true;
        }
        if (tokens.length && tokens[0].content == ',') {
            tokens.shift();
            continue;
        }
        throw new SchemeDefParseError(tokens[0], ['Komma', 'Unterstreichungsumschalter', 'schließende runde Klammer']);
    }
    throw new SchemeDefParseError(tokens[0], ['Attributname', 'Unterstreichungsumschalter', 'öffnende runde Klammer']);
}
