/** Array containing multiple table scheme defintions (SchemeDefTable instances), eventually connected by references */
class SchemeDefs extends Array{
    constructor() {
        super();
    }

    /**
    * Get contained SchemeDefTable instance having name `name`
    * @return SchemeDefTable        Requested table scheme definition
    */
    getTableByName(
        name        // Name of requested table
    ) {
        return this.reduce((prev, def) => prev || ((def.table == name) ? def : null), null);
    }

    /**
    * Get SchemeDefAttribute instance having name `name` of contained table having name `table`
    * @return SchemeDefAttribute    Requested attribute
    */
    getAttributeByName(
        table,      // Name of requested table
        name        // Name of requested attribute
    ) {
        const tableDef = this.getTableByName(table);
        return (tableDef instanceof SchemeDefTable) ? tableDef.getAttributeByName(name) : null;
    }

    /**
    * Get datatype of described attribute
    * @return mixed     String representing datatype or `undefined` if attribute exists but no datatype is declared or `NaN` if the attribute does not exist
    */
    getDatatype(
        table,      // Name of requested table
        name        // Name of requested attribute
    ) {
        const attribute = this.getAttributeByName(table, name);
        return (attribute instanceof SchemeDefAttribute) ? attribute.datatype : NaN;
    }

    /**
    * Check if all or not a single table declares data types of attributes
    * @return boolean       Do either all or none tables declare data types?
    */
    checkDatatypeDeclarationConsistency() {
        const hasDatatypes = this.map(def => def.hasDatatypes()).filter(value => (value !== undefined));
        return hasDatatypes.reduce((prev, curr) => prev && (curr == hasDatatypes[0]), true);
    }

    /**
    * Find name of first duplicate table
    * @return mixed         First found table name given to at least two tables
    */
    getDuplicateName() {
        const names = [];
        return this.reduce((prev, { table }) => {
            if (prev)
                return prev;
            if (names.indexOf(table) > -1)
                return table;
            names.push(table);
            return false;
        }, false);
    }

    /**
    * Check if table scheme definitions are semanticaly correct
    * @return mixed         Not falsey string describing first error or `false` (meaning no error)
    */
    validate(
        arrAllowedDatatypes     // Array of strings representing all allowed data types
    ) {
        const duplicate = this.getDuplicateName();
        if (duplicate)
            return 'Es kann nicht mehrere Definitionen der Tabelle „' + duplicate + '“ geben.';
        if (!this.checkDatatypeDeclarationConsistency())
            return 'Datentypen von Attributen müssen entweder in jedem oder in keinem Tabellenschema definiert werden.';
        return this.reduce((prev, def) => prev || def.validate(this, arrAllowedDatatypes), false);
    }

    /**
    * Calculate best fit of this table schemes and target data
    * @return Object    Objects like { match: "Fraction of match", desc: "Human readable text describing main difference" }
    */
    match(
        targetData      // Description of wanted structure beeing object like { keepOrder: "true if order of tables is relevant", content: "Array of arrays of objects describing tables" }
    ) {
        return SchemeDefs.matchGeneral(
            this,
            (item, target) => item.match(target),
            targetData,
            {
                internalTargetInvalid: 'Interner Fehler: Beschreibung der Soll-Tabellenschemata liegt in falschem Format vor',
                missingItem: target => 'Schema für ' + (target.table ? ('Tabelle ' + target.table) : 'eine Tabelle') + ' fehlt',
                surplusItem: item => 'Tabelle ' + item.table + ' ist überzählig',
                needMore: 'Es müsste mehr Tabellenschemata geben',
                needLess: 'Es müssten weniger Tabellenschemata sein'
            }
        );
    }
}

/**
* Read all table scheme definitions from tokens without changing `tokens`
* @return SchemeDefs        Array of SchemeDefTable instances
*/
SchemeDefs.parse = function(
    tokens      // Array of tokens (without whitespace!) to read from
) {
    const realTokens = tokens.filter(token => token.type != SchemeDefToken.TYPE_WHITESPACE);
    const defs = new SchemeDefs();
    while (realTokens.length) {
        if (!SchemeDefTable.parse(realTokens, defs))
            throw new SchemeDefParseError(realTokens[0], 'Tabellenname');
    }
    return defs;
}



/**
* Calculate best fit of objects contained by calling object and target data
* @return Object    Objects like { match: "Fraction of match", desc: "Human readable text describing main difference" }
*/
SchemeDefs.matchGeneral = function(
    arrContent,     // Array of objects contained by calling object
    matchItem,      // Function like (item, target) => { match, desc } calculating match of item and target values
    targetData,     // Description of wanted structure beeing object like { keepOrder: "true if order of tables is relevant", content: "Array of arrays of objects describing tables" }
    errorDescs      // Object containing and creating error descriptions
) {
    if (targetData instanceof Array)
        return SchemeDefs.matchGeneral.noOrder(arrContent, matchItem, targetData, errorDescs);
    if (!(targetData instanceof Object))
        return { match: 0, desc: errorDescs.internalTargetInvalid };
    if (targetData.keepOrder)
        return SchemeDefs.matchGeneral.keepOrder(arrContent, matchItem, targetData.content, errorDescs);
    return SchemeDefs.matchGeneral.noOrder(arrContent, matchItem, targetData.content, errorDescs);
}

/**
* Calculate best fit of obects contained by calling object and target data assuming fixed order of table schemes
* @return Object    Objects like { match: "Fraction of match", desc: "Human readable text describing main difference" }
*/
SchemeDefs.matchGeneral.keepOrder = function(
    arrContent,     // Array of objects contained by calling object
    matchItem,      // Function like (item, target) => { match, desc } calculating match of item and target values
    targetData,     // Description of wanted structure beeing object like { keepOrder: "true if order of tables is relevant", content: "Array of arrays of objects describing tables" }
    errorDescs      // Object containing and creating error descriptions
) {
    if (!(targetData instanceof Array))
        return { match: 1, desc: '' };
    if (!targetData.length || !(targetData[0] instanceof Array))
        targetData = [ targetData ];
    return targetData.map(target => {
        const itemMatches = target.map((targetTable, index) => {
            if (index >= arrContent.length)
                return { match: 0, desc: errorDescs.missingItem(targetTable) };
            return matchItem(arrContent[index], targetTable);
        });
        if (arrContent.length > target.length)
            arrContent.slice(target.length).forEach(def => itemMatches.push({ match: 0, desc: errorDescs.surplusItem(def) }));
        if (!itemMatches.length)
            return { match: 1, desc: '' };
        return itemMatches.reduce((prev, curr) => ({ match: prev.match + curr.match / itemMatches.length, desc: prev.desc || curr.desc }), { match: 0, desc: '' });
    }).reduce((prev, curr) => (prev.match > curr.match) ? prev : curr, { match: 0, desc: errorDescs.internalTargetInvalid });
}

/**
* Calculate best fit of this table schemes and target data assuming order of table schemes is irrelevant
* @return Object    Objects like { match: "Fraction of match", desc: "Human readable text describing main difference" }
*/
SchemeDefs.matchGeneral.noOrder = function(
    arrContent,     // Array of objects contained by calling object
    matchItem,      // Function like (item, target) => { match, desc } calculating match of item and target values
    targetData,     // Description of wanted structure beeing object like { keepOrder: "true if order of tables is relevant", content: "Array of arrays of objects describing tables" }
    errorDescs      // Object containing and creating error descriptions
) {
    if (!(targetData instanceof Array))
        return { match: 1, desc: '' };
    if (!targetData.length || !(targetData[0] instanceof Array))
        targetData = [ targetData ];
    return targetData.map(target => {
        if (target.length != arrContent.length)
            return {
                match: Math.min(target.length, arrContent.length) / Math.max(target.length, arrContent.length),
                desc: (target.length > arrContent.length) ? errorDescs.needMore : errorDescs.needLess
            };
        const itemQualifications = arrContent.map(def => target.map(targetTable => matchItem(def, targetTable)));
        let minMatchMatches = Infinity;
        let minMatchThisIndex = -1;
        let minMatchTargetIndex = -1;
        while (itemQualifications.length) {
            const itemMatches = itemQualifications.map(qualifications => qualifications.filter(obj => (obj.match == 1)).length);
            minMatchMatches = itemMatches.reduce((prev, curr) => (curr && curr < prev) ? curr : prev, Infinity);
            if (!Number.isFinite(minMatchMatches))
                break;
            minMatchThisIndex = itemMatches.reduce((prev, curr, index) => (curr == minMatchMatches && prev == -1) ? index : prev, -1);
            minMatchTargetIndex = itemQualifications[minMatchThisIndex].reduce((prev, curr, index) => (curr.match == 1 && prev == -1) ? index : prev, -1);
            itemQualifications.splice(minMatchThisIndex, 1);
            itemQualifications.forEach(q => q.splice(minMatchTargetIndex, 1));
        }
        if (!itemQualifications.length)
            return { match: 1, desc: '' };
        let fraction = arrContent.length - itemQualifications.length;
        let desc = '';
        while (itemQualifications.length) {
            const bestMatch = itemQualifications.map(qualifications => qualifications.reduce((prev, curr) => Math.max(prev, curr.match), 0));
            minMatchMatches = bestMatch.reduce((prev, curr) => Math.max(prev, curr));
            minMatchThisIndex = bestMatch.reduce((prev, curr, index) => (curr == minMatchMatches && prev == -1) ? index : prev, -1);
            minMatchTargetIndex = itemQualifications[minMatchThisIndex].reduce((prev, curr, index) => (curr.match == minMatchMatches && prev == -1) ? index : prev, -1);
            desc = desc || itemQualifications[minMatchThisIndex][minMatchTargetIndex].desc;
            fraction += minMatchMatches;
            itemQualifications.splice(minMatchThisIndex, 1);
            itemQualifications.forEach(q => q.splice(minMatchTargetIndex, 1));
        }
        return { match: fraction / arrContent.length, desc };
    }).reduce((prev, curr) => (prev.match > curr.match) ? prev : curr, { match: 0, desc: errorDescs.internalTargetInvalid });
}
